<?php

namespace Internetgalerie\IgDatapoolFe\Services\Exporter;

/**
 * This class provides an export
 */
abstract class ExporterService
{
    /**
     * @var \TYPO3\CMS\Extbase\Reflection\ReflectionService
     */
    protected $reflectionService;

    protected $fields = [];
    protected $headers = [];
    protected $options = ['filename' => ''];

    // conversion options for each field. e.g., $fieldOptions['fieldname']['format'] = 'd.m';
    protected $fieldOptions = [];

    // caches the converted lines for later use. used by JSON and XML
    protected $res = [];

    // if this is set, a function is used to get the export row insead of the array
    protected $exporterFunction = null;

    /**
     * @param \TYPO3\CMS\Extbase\Reflection\ReflectionService $reflectionService
     */
    public function injectReflectionService(\TYPO3\CMS\Extbase\Reflection\ReflectionService $reflectionService)
    {
        $this->reflectionService = $reflectionService;
    }

    /**
     * Set the options for export
     *
     * @param string $type
     * @param mixed  $fields  string (method) or array of field=>label pairs
     * @param array  $options
     */
    public function setOptions(array $fields, $options = [])
    {
        if (is_string($fields)) {
            $this -> exporterFunction = $fields;
        } elseif (is_array($fields) && count($fields)) {
            foreach ($fields as $key => $value) {
                if (is_numeric($key)) {
                    if ($options['dbLabel']) {
                        $this -> headers[$value] = \Internetgalerie\IgDatapoolFe\Utility\NeedfulThings::getFieldLabelByProperty($value);
                    } else {
                        $this -> headers[$value] = $value;
                    }
                    $this -> fields[] = $value;
                } else {
                    $this -> fields[] = $key;
                    $this -> headers[$key] = $value;
                }
            }
        } else {
            $this -> fields = $this -> getProperties();
        }

        foreach ($options as $key => $val) {
            $this -> options[$key] = $val;
        }
        if (!$this -> options['filename']) {
            $this -> options['filename'] = \Internetgalerie\IgDatapoolFe\Controller\ActionController::$currentController -> getRequest() -> getControllerName() . 's';
        }
    }

    /**
     * exports a list of models
     *
     * @var mixed \Iterator<\TYPO3\CMS\Extbase\DomainObject\AbstractEntity>
     */
    public function export($models)
    {
        $this -> before();
        foreach ($models as $key => $model) {
            if ($this -> exporterFunction !== null) {
                if (!method_exists($model, $this -> exporterFunction)) {
                    throw new \Exception('Model ' . get_class($model) . ' hat keine entsprechende Methode: ' . $this -> exporterFunction);
                }
                $line = $model -> {$this->exporterFunction}();
                foreach ($line as $field => &$value) {
                    $value = $this -> convertToString($value);
                }
                $this -> exportArray($line, $key);
            } elseif (is_array($model)) {
                $this -> exportArray($model, $key);
            } elseif ($model instanceof \TYPO3\CMS\Extbase\DomainObject\AbstractEntity) {
                $this -> exportModel($model);
            }
        }
        $this -> after();
        exit ;
    }

    /**
     * Export by method
     */
    public function exportByMethod($array)
    {
        $res = [];
        foreach ($this->fields as $f) {
            $res[$f] = $this -> getString($f, $model);
        }
        $uid = uniqid();
        if (method_exists($model, 'getUid')) {
            $uid = $model -> getUid();
        }
        $this -> res[$uid] = $res;
        $this -> exportLine($res, $uid);
    }

    /**
     * Simple export for given arrays
     */
    public function exportArray(array $arr, $key)
    {
        $this -> exportLine($arr, $key);
        $this -> res[$key] = $arr;
    }

    /**
     * export a single model
     *
     * @param \TYPO3\CMS\Extbase\DomainObject\AbstractEntity $model
     */
    public function exportModel(\TYPO3\CMS\Extbase\DomainObject\AbstractEntity $model)
    {
        $res = [];
        foreach ($this->fields as $f) {
            $res[$f] = $this -> getString($f, $model);
        }
        $uid = uniqid();
        if (method_exists($model, 'getUid')) {
            $uid = $model -> getUid();
        }
        $this -> res[$uid] = $res;
        $this -> exportLine($res, $uid);
    }

    abstract public function exportLine($array, $uid);

    abstract public function before();

    abstract public function after();

    protected function getString($path, $object)
    {
        if ($path === 'uid') {
            return $object -> getUid();
        }
        $val = \TYPO3\CMS\Extbase\Reflection\ObjectAccess::getPropertyPath($object, $path);
        return $this -> convertToString($val);
    }

    /**
     * Converts whatever to a string.
     * can have options, @see $this->fieldOptions[]
     */
    protected function convertToString($val)
    {
        $str = '';
        if ($val === null) {
            $str = '';
        } elseif (is_object($val)) {
            switch (get_class($val)) {
            case 'DateTime':
                // FIXME(mb): does this work with recursion, when ObjectStorage is used!?
                $format = isset($this -> fieldOptions[$path]['format']) ? $this -> fieldOptions[$path]['format'] : 'Y-m-d';
                $str = $val -> format($format);
                break;
            case 'Internetgalerie\IgDatapoolFe\Domain\Model\FileReference':
            case 'TYPO3\CMS\Extbase\Domain\Model\FileReference':
                if ($val -> getOriginalResource()) {
                    $str = $val -> getOriginalResource() -> getOriginalFile() -> getPublicUrl();
                } else {
                    $str = '';
                }
                break;
            case 'TYPO3\CMS\Extbase\Persistence\ObjectStorage':
                if (count($val) > 0) {
                    $key = 0;
                    $str = $this -> getString($key, $val);
                } else {
                    $str = '';
                }
                break;
            default:
                $str = (string)$val;
                break;
            }
        } else {
            $str = (string)$val;
        }
        return $str;
    }

    /**
     * Gets the properties of an object
     *
     * @return array properety => type
     */
    protected function getProperties()
    {
        $class = \Internetgalerie\IgDatapoolFe\Controller\ActionController::$currentObjectClass;
        $properties = $this -> reflectionService -> getClassPropertyNames($class);
        //filter system fields
        $system = ['pid', '_localizedUid', '_languageUid', '_versionedUid'];
        $properties = array_diff($properties, $system);
        return $properties;
    }

    /**
     * Sets an array of headers.
     *
     * @param array $headers field => header pairs for the headers
     */
    public function setHeaders(array $headers)
    {
        $this -> headers = $headers;
    }

    /**
     * Returns all headers for each field
     *
     * @param string $fieldname return only one label for a field
     */
    protected function getHeaders($fieldname = false)
    {
        if ($fieldname) {
            return $this -> headers[$fieldname];
        }
        return $this -> headers;
    }

    /**
     * Sets an array of field options
     *
     * @param array $fieldOptions
     */
    public function setFieldOptions(array $fieldOptions)
    {
        $this -> fieldOptions = $fieldOptions;
    }
}
